#!/usr/local/bin/perl

# $Id: CDBurning.pm,v 1.11 2004/01/02 06:58:44 foley Exp foley $
##############################################################################
#                                                                            #
#           Copyright by (c) Regents of the University of California         #
#                      University of California, San Diego                   #
#                      Scripps Institution of Oceanography                   #
#                           Shipboard Computer Group                         #
#                       La Jolla, California, USA 92093                      #
#                                                                            #
#                          All rights reserved. 2002                         #
#                                                                            #
#----------------------------------------------------------------------------#
#                                                                            #
#                                CDBurning.pm                                #
#                                                                            #
# Written By: Bill French (french@sdsioa.ucsd.edu)                           #
# Created:    Thu Nov 21 19:44:20 GMT 2002                                   #
#                                                                            #
# Purpose:                                                                   #
#                                                                            #
#   Provide functions to create and burn CDs to the cdrw server.  The        #
#   ISO9660 image is created on the local machine and sent to the cdrw       #
#   server through a ssh pipe.  We log on to the cdrw server as a restricted #
#   cdrw user and call cdrecord once the ISO image is created.               #
#                                                                            #
# Exported Functions:                                                        #
#                                                                            #
#   make_iso  --  Create an iso9660 image and transfer it to the cdrw server.#
#                                                                            #
#   write_cd  --  Start the burn process on the cdrw server.                 #
#                                                                            #
#   update_sequence_number  --  Update the CD sequence number.               #
#                                                                            #
# Internal Functions:                                                        #
#                                                                            #
#   check_file_sizes     --  Check the total size of the source files to     #
#                            ensure that everything will fit on one CD.      #
#                                                                            #
#   get_ssh_key_file     --  Take ownership of the ssh used to logon to the  #
#                            cdrw server.                                    #
#                                                                            #
#   clean_ssh_key_file   --  Destroy the ssh key that we took control of     #
#                            with get_ssh_key_file.                          #
#                                                                            #
#   get_sequence_number  --  return the current sequence number for the next #
#                            CD to burn.                                     #
#                                                                            #
#   is_cd_mounted        --  is the local cd drive mounted?                  #
#                                                                            #
# Global Variables:                                                          #
#                                                                            #
#   $VARDIR  --  This is the directory name where we keep persistant data.   #
#                                                                            #
#   $LOCAL_CD_MOUNT_POINT  --  This is the mount point of the CD Rom on this #
#                              system.                                       #
#                                                                            #
# Files:                                                                     #
#                                                                            #
#   $VARDIR/copyrght.file  --  The copyright file                            #
#                                                                            #
#   $VARDIR/Sequence_no    --  Storage for the next sequence number          #
#                                                                            #
#   $VARDIR/cdrw-key       --  ssh key to log on to the cdrw server          #
#                                                                            #
# Additional Notes:                                                          #
#                                                                            #
#   This script uses ssh to talk to the cdrw system.  In order to work       #
#   without passwords we use ssh keys.  To create these keys do the          #
#   following.                                                               #
#                                                                            #
#      In this source directory:                                             #
#                                                                            #
#         $ /bin/ssh-keygen -f cdrw-key -t rsa                               #
#         $ ./install                                                        #
#                                                                            #
#      On the CDRW server:                                                   #
#                                                                            #
#         Make sure that the key in cdrw-key.pub is in                       #
#         ~cdrw/.ssh/authorized_keys file.                                   #
#                                                                            #
# Future Upgrades:                                                           #
#                                                                            #
#   I would like to have the copyright file generated so the date would      #
#   always be correct.                                                       #
#                                                                            #
#   There could be possile race conditions.  File locking for the iso image  #
#   on the remote server would be a good thing.                              #
#                                                                            #
##############################################################################


use lib ('/usr/local/lib/perl5');
package Scg::CDBurning;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw( is_cd_mounted
              make_iso 
              update_sequence_number
              write_cd
              $LOCAL_CD_MOUNT_POINT
              $VARDIR );

#
#  Constants
#

$MAX_DVD_SIZE         = 699999;    # How much data can we put on a CD?  Don't
                                  # forget we need space for translation
                                  # tables and FAT info.

$MAX_SEQUENCE_NUMBER = 9999;      # What is the max sequence number before we
                                  # roll it back to 1.


# Directory for all of out persistant data (i.e. sequence_number)
$VARDIR           = "/SCG.PROC/sproc/cd_rom_scripts";

$COPYRIGHT_FILE   = "$VARDIR/copyrght.fil";  # Template copyright file
$SEQUENCE_NO_FILE = "$VARDIR/Sequence_no";   # The next sequence number

$WRITTER_SERVER   = "mlv-cdrw";	             # CDRW Remote server
$WRITTER_USER     = "cdrw";                  # Login to the CDRW server as:
$WRITTER_KEY      = "$VARDIR/cdrw-key";      # SSH Private key for the server
$WRITTER_DEST     = "/var/images/image.iso"; # Where to put the image on the 
                                             # server.

$MKISO_LOG        = '/tmp/mkiso.log';        # Log file for the mkisofs command

$CDRECORD_SPEED   = '52';     # Write speed of the cdrw drive 
$CDRECORD_DEVICE  = '0,0,0';  # The device address. format: scsibus,target,lun

$LOCAL_CD_MOUNT_POINT = '/cdrom/cdrom0'; # The mount point of the cdrom on this
                                         # machine.  This is not the CDRW.

$MOUNT_TIMEOUT        = 30;              # How many seconds to wait before we
                                         # timeout waiting for the CD to mount.

$| = 1;   # unbuffer output

#
# Utilities
#

# Local Utilities
$AWK   = '/usr/bin/awk';
$CP    = '/bin/cp';
$FIND  = '/usr/local/bin/find';
$MKISO = '/usr/local/bin/mkisofs';
$RSH   = '/usr/bin/ssh';

# Remote machine Utilities
$CAT      = '/bin/cat';
$CDRECORD = '/usr/bin/cdrecord';


#
#  Function Definitions
#

##############################################################################
# Function: make_iso                                                         #
#                                                                            #
# Create an ISO9660 images with the file list passed to the function if the  #
# size of the image will fit on one CD.  The image is piped through an ssh   #
# tunnel as it is created.  If everything works out ok 1 is returned.  This  #
# would indicate that it is ok to call the write_cd function.                #
#                                                                            #
# This function takes a lists that contains the source list of files to burn.#
# It will accept both files and directories in the list.  If the list item   #
# is a directory then the entire contents of the directory will be added to  #
# the iso.                                                                   #
#                                                                            #
# All source files will be burned to the root of the cdrom.  To select a     #
# different directory to install the files on, use the format:               #
#                                                                            #
#    'cd_subdir/=/path/to/source'                                            #
#                                                                            #
# This will install the contents of source in a subdirectory on the CD       #
# called cd_subdir.                                                          #
#                                                                            #
# The same can be done for files.  ie:                                       #
#                                                                            #
#    'dest_dir/filename=/path/to/source/file'                                #
#                                                                            #
# This will install the file named 'file' on the CD in a subdirectory called #
# dest_dir and the file will be called 'filename'.                           #
#                                                                            #
# For more information about this see grafting in the mkisofs manpage.       #
#                                                                            #
# Parameters: @target_list.                                                  #
#                                                                            #
#   target_list  --  This contains a list of all the source paths to include #
#		     in the ISO image.  All paths specified will be place in #
#                    the root directory of the image unless the method above #
#                    is used.                                                #
#                                                                            #
# Return Values:                                                             #
#                                                                            #
#   1  --  We have created and transfered the image with no problems.        #
#   0  --  The data specified was too large to fit on one CD.                #
#                                                                            #
##############################################################################

sub make_iso
{
   use lib ('/usr/local/lib/perl5');
   use Scg::Cruiseinfo;

   my @target_list = @_;

   # Take ownership of the ssh key
   my $tmpkey = get_ssh_key_file( $WRITTER_KEY );

   # Get the sequence number, if one can't be found use $MAX_SEQUENCE_NUMBER
   my ($seqno) = &get_sequence_number || $MAX_SEQUENCE_NUMBER;

   # Set the CD volume label to the Cruisename_Sequence-number if the Cruise
   # name is set, otherwise use SIO_Sequence-number.
   my ($vol_label) = $CRUISE_INFO{'CRUISE'} ? "$CRUISE_INFO{'CRUISE'}_$seqno" :
                                              "SIO_$seqno";

   # Make sure that we have files to burn.
   @target_list eq 0
      && die( "No targets specified! " );

   # Ensure that the data is not too large for one CD.
   unless( check_file_sizes( $MAX_DVD_SIZE, @target_list ) ) {
      return 0;
   }

   print "\nBuilding the ISO9660 image and transfering it to the burner.\n";
   print "This could take several minutes, please be patient.\n";

   # Command to log on to the remote server.
   my $rsh_command = "$RSH -i $tmpkey $WRITTER_USER\@$WRITTER_SERVER";

   # The command to run on the remote server.  Basically take STDIN and send
   # it to a temp iso file.
   my $remote_command = "$CAT > $WRITTER_DEST ";

   # Command to create the ISO image and transfer it to the cdrw server
   my $command = "$MKISO -d -L -r -T -J -V $vol_label " . 
		 "-graft-points -log-file $MKISO_LOG " .
                 "-o /dev/stdout @target_list $COPYRIGHT_FILE "  . 
                 "| $rsh_command \"$remote_command\"";

   # Run the command. 
   $res = system $command;

   # Release the ssh key
   &clean_ssh_key_file;

   # If the command was not successful then print an error and return failure.
   if( $res ne 0 )
   {
      print STDERR "Failed to create ISO9660 Image\n";
      return 0;
   }

   print "\nISO9660 image create, ";

   # Clean up the log because everything went ok.
#   unlink $MKISO_LOG;

   return 1;
}


##############################################################################
# Function: write_cd                                                         #
#                                                                            #
# Write an image on the cdrw server to CD.  This function will ssh to the    #
# cdrw server and use cdrecord to burn the temporary image.                  #
#                                                                            #
# This function could update the sequence number, but I choose not to do     #
# that so we can burn multiple copies of the CD.  If the burn is successful  #
# then the calling function should also make a call to                       #
# &update_sequence_number.                                                   #
#                                                                            #
# Parameters: none                                                           #
#                                                                            #
# Return Values:                                                             #
#                                                                            #
#   1  --  We have burned the CD with no problems.                           #
#   0  --  There was some failure.                                           #
#                                                                            #
##############################################################################

sub write_cd
{
   # Stop ctrl-c (really fucks up the scsi channel if intrupted)
   $SIG{INT} = 'IGNORE';

   # Take ownership of the key
   my $tmpkey = get_ssh_key_file( $WRITTER_KEY );

   # Remote shell command
   my $rsh_command = "$RSH -i $tmpkey $WRITTER_USER\@$WRITTER_SERVER";

   # Command to burn the CD
   my $remote_command = "$CDRECORD -v -dao -eject " .
                                  "speed=$CDRECORD_SPEED " .
                                  "dev=$CDRECORD_DEVICE " .
                                  "-data $WRITTER_DEST";

   print "now begining the burn process.\n\n";

   # Burn the CD
   $res = system "$rsh_command \"$remote_command\"";
                
   # Release the CD key
   &clean_ssh_key_file;

   # Did we burn everything correctly?
   unless( $res eq 0 )
   {
      print STDERR "Failed to burn the CD!\n";
      return 0;
   }

   return 1;
}    


##############################################################################
# Function: check_files_sizes                                                #
#                                                                            #
# This function serves two purposes, it ensures that all of the paths exist  #
# and that their total size is less then the MAX_DVD_SIZE.                    #
#                                                                            #
# We do a find -ls to get the file sizes, then we total them all together.   #
# I chose to use find/awk vs. du to get the file sizes because du rounds up  #
# to the nearest byte.  This would get us an less accurate file size reports.#
#                                                                            #
# Parameters: $max_size, @files                                              #
#                                                                            #
#   max_size  --  The maximum size that we will allow per CD (in kb);        #
#                                                                            #
#   files     --  List containing paths that are going to be recorded.       #
#                 These paths can take grafting information just like the    #
#                 list passwd to &make_iso.  See that documentation for more #
#                 infomation about this.                                     #
#                                                                            #
# Return Values:                                                             #
#                                                                            #
#   1  --  All files exist and they will fit on one CD.                      #
#   0  --  Too big for a CD.                                                 #
#                                                                            #
##############################################################################

sub check_file_sizes
{
   my $max_size = shift;
   my @files = @_;

   my $total_size = 0;

   print "Calculating file sizes\n\n";

   # Clean up the path
   foreach $file (@files)
   {
      # If this is a grafted path then remove the grafting information and 
      # you are just left with the source path.
      $target_file = $file =~ /.*=.*/ ? (split( /=/, $file ))[1] : $file;

      # Ensure that the path exists and is readable by you!
      die( "Could not open input file: $target_file" ) unless( -r $target_file );

      # Uses find and awk to find the total size used by the file or directory 
      # in bytes.
      $size = `$FIND $target_file -ls | $AWK '{ sum += \$7 } END { printf( "%d", sum ) }'`;

      # Display the size in Kb and the path
      printf( "%-12d%s\n", $size / 1024, $target_file );

      $total_size += $size;
   }

   # Display the total size in MB.
   print "\nTotal size: " . int($total_size / 1048576) . " MB\n";

   # Are we under the size limit?
   if($total_size / 1024 > $max_size) {
      print "Total size is larger the the maximum storage capacity.\n\n";
      return 0;
   }

   return 1;
}


##############################################################################
# Function: get_ssh_key_file                                                 #
#                                                                            #
# Take control of an ssh key file.  Basically we copy an ssh key file to a   #
# temp file, with a umask of 077, and then return the filename.              #
#                                                                            #
# Parameters: none                                                           #
#                                                                            #
# Return Values:                                                             #
#                                                                            #
#   Return the filename of the new key file.                                 #
#                                                                            #
##############################################################################

sub get_ssh_key_file
{
   my $key_file = shift;
   my $tmpfile = "/tmp/tmpfile.$$";

   $old_mask = umask 077;

   system "$CP $key_file $tmpfile";

   umask $old_mask;

   return $tmpfile;
}


##############################################################################
# Function: clean_ssh_key_file                                               #
#                                                                            #
# Delete the ssh key file that we created with get_ssh_key_file.  This       #
# function is called when we see a SIGQUIT message.                          #
#                                                                            #
# Parameters: none                                                           #
#                                                                            #
# Return Values: none                                                        #
#                                                                            #
##############################################################################

sub clean_ssh_key_file
{
   my $keyfile = "/tmp/tmpfile.$$";

   unlink $keyfile;
}

##############################################################################
# Function: get_sequence_number                                              #
#                                                                            #
# Return the sequence number that is stored in the $SEQUENCE_NO_FILE.  If    #
# we are unable to open the file then return 0;                              #
#                                                                            #
# Parameters: none                                                           #
#                                                                            #
# Return Values:                                                             #
#                                                                            #
#    Return the number contained in the $SEQUENCE_NO_FILE or 0 if we could   #
#    not open the file.                                                      #
#                                                                            #
##############################################################################

sub get_sequence_number
{
   open( SEQ_FILE, "$SEQUENCE_NO_FILE" )
      || return 0;

   my ($seqno) = <SEQ_FILE>;
   chomp $seqno;
   close SEQ_FILE;

   return $seqno;
}


##############################################################################
# Function: update_sequence_number                                           #
#                                                                            #
# Incriment the sequence number stored in $SEQUENCE_NO_FILE.  If we couldn't #
# read or write to the file then output a warning message.                   #
#                                                                            #
# Parameters: none                                                           #
#                                                                            #
# Return Values:  none                                                       #
#                                                                            #
##############################################################################

sub update_sequence_number
{
   my ($seqno) = &get_sequence_number + 1 || -1;

   # If the sequence number is too large then roll it to 1
   $seqno = $seqno > $MAX_SEQUENCE_NUMBER ? 1 : $seqno;

   if( $seqno eq -1 or open( SEQ_FILE, ">$SEQUENCE_NO_FILE" ) )
   {
      print SEQ_FILE $seqno . "\n";
      close SEQ_FILE;
   }
   else {
      print STDERR "Couldn't write sequence number file: $SEQUENCE_NO_FILE\n";
   }
}

##############################################################################
# Function: is_cd_mounted                                                    #
#                                                                            #
# Return if the local cdrom has mounted.  The funtion will keep checking if  #
# the drive has mounted until a timeout occures.                             #
#                                                                            #
# Parameters: none                                                           #
#                                                                            #
# Return Values:                                                             #
#                                                                            #
# Return 1 if we can get a file list from the CD, otherwise return 0 if we   #
# have exceeded out timeout and no files we found.                           #
#                                                                            #
##############################################################################

sub is_cd_mounted
{
   my $start_time = time;

   # Get the file list if the mount point exists, otherwise it is an empty list
   my @file_list = (-e $LOCAL_CD_MOUNT_POINT) ? <$LOCAL_CD_MOUNT_POINT/*> : ();

   # If we don't alreay have a file list then we will keep trying
   if ( @file_list == 0 ) { print "Waiting for CD to mount ."; }

   # Keep checking the mount point for files until we exceed out timeout value
   while( $start_time + $MOUNT_TIMEOUT > time && @file_list <= 0 )
   {
      @file_list = (-e $LOCAL_CD_MOUNT_POINT) ? <$LOCAL_CD_MOUNT_POINT/*> : ();
      print ".";
      sleep 1;
   }

   print "\n";

   # If we still dont have a list of files after the timeout then give up
   unless( @file_list > 0 )
   {
      print STDERR "CD Mount timeout, No CD mounted.\n";
      return 0;
   }

   return 1;
}

